### EXAMPLE 2: demonstrates how measurement error can lead to a type sign error in a covariate
### Even when you have a good predictor, if it's biased against a covariate you can get the wrong sign.
### Even when you include the proxy variable in the regression.
### But with some ground truth and multiple imputation, you can fix it.

library(argparser)
library(mecor)
library(ggplot2)
library(data.table)
library(filelock)
library(arrow)
library(Amelia)
library(Zelig)
library(predictionError)
options(amelia.parallel="no",
        amelia.ncpus=1)

## SETUP:
### we want to estimate x -> y; x is MAR
### we have x -> k; k -> w; x -> w is used to predict x via the model w.
### A realistic scenario is that we have an NLP model predicting something like "racial harassment" in social media comments
### The labels x are binary, but the model provides a continuous predictor

### simulation:
#### how much power do we get from the model in the first place? (sweeping N and m)
#### 
source("simulation_base.R")

simulate_latent_cocause <- function(N, m, B0, Bxy, Bgy, Bkx, Bgx, seed){
    set.seed(seed)

    ## the true value of x

    g <- rbinom(N, 1, 0.5)
    k <- rnorm(N, 0, 1)
    xprime <- Bkx*k + Bgx * g + rnorm(N,0,1)
    xvec <- scale(xprime)

    y <-  Bxy * xvec  + Bgy * g + rnorm(N, 0, 1) + B0

    df <- data.table(x=xvec,k=k,y=y,g=g)
    names(df) <- c('x','k','y','g')
    if( m < N){
        df <- df[sample(nrow(df), m), x.obs := x]
    } else {
        df <- df[, x.obs := x]
    }

    w.model <- lm(x ~ k,df)
    w <- predict(w.model,data.frame(k=k)) 
    w <- logistic(w + rnorm(N,0,sd(w)*0.1))
    ## y = B0 + B1x + e

    df[,':='(w=w, w_pred = as.integer(w>0.5))]
    return(df)
}


parser <- arg_parser("Simulate data and fit corrected models")
parser <- add_argument(parser, "--N", default=1000, help="number of observations of w")
parser <- add_argument(parser, "--m", default=200, help="m the number of ground truth observations")
parser <- add_argument(parser, "--seed", default=4321, help='seed for the rng')
parser <- add_argument(parser, "--outfile", help='output file', default='example_2.feather')
args <- parse_args(parser)

Ns <- c(1000, 10000, 1e6)
ms <- c(100, 250, 500, 1000)
B0 <- 0
Bxy <- 0.2
Bgy <- -0.2
Bkx <- 2
Bgx <- 3

outline <- run_simulation(simulate_latent_cocause(args$N, args$m, B0, Bxy, Bgy, Bkx, Bgx, args$seed)
                         ,list('N'=args$N,'m'=args$m,'B0'=B0,'Bxy'=Bxy,'Bgy'=Bgy, 'Bkx'=Bkx, 'Bgx'=Bgx, 'seed'=args$seed))

outfile_lock <- lock(paste0(args$outfile, '_lock'),exclusive=TRUE)
if(file.exists(args$outfile)){
    logdata <- read_feather(args$outfile)
    logdata <- rbind(logdata,as.data.table(outline))
} else {
    logdata <- as.data.table(outline)
}

print(outline)
write_feather(logdata, args$outfile)
unlock(outfile_lock)
